# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from datetime import datetime

from breezy.plugins.gitlab.forge import (
    NotGitLabUrl,
    NotMergeRequestUrl,
    parse_gitlab_merge_request_url,
    parse_timestring,
)
from breezy.tests import TestCase


class ParseGitLabMergeRequestUrlTests(TestCase):
    def test_invalid(self):
        self.assertRaises(
            NotMergeRequestUrl,
            parse_gitlab_merge_request_url,
            "https://salsa.debian.org/",
        )
        self.assertRaises(
            NotGitLabUrl, parse_gitlab_merge_request_url, "bzr://salsa.debian.org/"
        )
        self.assertRaises(
            NotGitLabUrl, parse_gitlab_merge_request_url, "https:///salsa.debian.org/"
        )
        self.assertRaises(
            NotMergeRequestUrl,
            parse_gitlab_merge_request_url,
            "https://salsa.debian.org/jelmer/salsa",
        )

    def test_old_style(self):
        self.assertEqual(
            ("salsa.debian.org", "jelmer/salsa", 4),
            parse_gitlab_merge_request_url(
                "https://salsa.debian.org/jelmer/salsa/merge_requests/4"
            ),
        )

    def test_new_style(self):
        self.assertEqual(
            ("salsa.debian.org", "jelmer/salsa", 4),
            parse_gitlab_merge_request_url(
                "https://salsa.debian.org/jelmer/salsa/-/merge_requests/4"
            ),
        )


class ParseTimestringTests(TestCase):
    def test_simple(self):
        self.assertEqual(
            datetime(2018, 9, 7, 11, 16, 17, 520000),
            parse_timestring("2018-09-07T11:16:17.520Z"),
        )



# Copyright (C) 2006 Canonical Ltd
# Copyright (C) 2008 Aaron Bentley <aaron@aaronbentley.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os

from breezy.errors import BinaryFile
from breezy.iterablefile import IterableFile
from breezy.patch import PatchInvokeError, diff3, iter_patched_from_hunks, run_patch
from breezy.patches import parse_patch
from breezy.tests import TestCase, TestCaseInTempDir


class TestPatch(TestCaseInTempDir):
    def test_diff3_binaries(self):
        with open("this", "wb") as f:
            f.write(b"a")
        with open("other", "wb") as f:
            f.write(b"a")
        with open("base", "wb") as f:
            f.write(b"\x00")
        self.assertRaises(BinaryFile, diff3, "unused", "this", "other", "base")

    def test_missing_patch(self):
        self.assertRaises(
            PatchInvokeError, run_patch, ".", [], _patch_cmd="/unlikely/to/exist"
        )


class PatchesTester(TestCase):
    def datafile(self, filename):
        data_path = os.path.join(
            os.path.dirname(__file__), "test_patches_data", filename
        )
        return open(data_path, "rb")

    def data_lines(self, filename):
        with self.datafile(filename) as datafile:
            return datafile.readlines()

    def test_iter_patched_from_hunks(self):
        """Test a few patch files, and make sure they work."""
        files = [
            ("diff-2", "orig-2", "mod-2"),
            ("diff-3", "orig-3", "mod-3"),
            ("diff-4", "orig-4", "mod-4"),
            ("diff-5", "orig-5", "mod-5"),
            ("diff-6", "orig-6", "mod-6"),
            ("diff-7", "orig-7", "mod-7"),
        ]
        for diff, orig, mod in files:
            parsed = parse_patch(self.datafile(diff))
            orig_lines = list(self.datafile(orig))
            mod_lines = list(self.datafile(mod))
            iter_patched = iter_patched_from_hunks(orig_lines, parsed.hunks)
            patched_file = IterableFile(iter_patched)
            count = 0
            for patch_line in patched_file:
                self.assertEqual(
                    patch_line, mod_lines[count], "for file {}".format(diff)
                )
                count += 1
            self.assertEqual(count, len(mod_lines))



# Copyright (C) 2006, 2007, 2009, 2010 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import bedding, ignores, tests
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestIsIgnored(TestCaseWithWorkingTree):
    def setUp(self):
        super().setUp()
        if self.workingtree_format.ignore_filename != ".bzrignore":
            raise tests.TestNotApplicable(
                "format does not use .bzrignore for ignore patterns"
            )

    def _set_user_ignore_content(self, ignores):
        """Create user ignore file and set its content to ignores."""
        bedding.ensure_config_dir_exists()
        user_ignore_file = bedding.user_ignore_config_path()
        with open(user_ignore_file, "wb") as f:
            f.write(ignores)

    def test_is_ignored(self):
        tree = self.make_branch_and_tree(".")
        # this will break if a tree changes the ignored format. That is fine
        # because at the moment tree format is orthogonal to user data, and
        # .bzrignore is user data so must not be changed by a tree format.
        self.build_tree_contents(
            [
                (
                    ".bzrignore",
                    b"./rootdir\n"
                    b"randomfile*\n"
                    b"*bar\n"
                    b"!bazbar\n"
                    b"?foo\n"
                    b"*.~*\n"
                    b"dir1/*f1\n"
                    b"dir1/?f2\n"
                    b"RE:dir2/.*\\.wombat\n"
                    b"path/from/ro?t\n"
                    b"**/piffle.py\n"
                    b"!b/piffle.py\n"
                    b"unicode\xc2\xb5\n"  # u'\xb5'.encode('utf8')
                    b"dos\r\n"
                    b"\n"  # empty line
                    b"#comment\n"
                    b" xx \n",  # whitespace
                )
            ]
        )
        # We set user ignore file to contain '' to avoid patterns from
        # user ignore being used instead of bzrignore. For .e.g. If we
        # don't do this 'foo.~1~' will match '*~' default user ignore
        # pattern instead of '*.~*' from bzr ignore as we expect below.
        self._set_user_ignore_content(b"")
        # is_ignored returns the matching ignore regex when a path is ignored.
        # we check some expected matches for each rule, and one or more
        # relevant not-matches that look plausible as cases for bugs.
        self.assertEqual("./rootdir", tree.is_ignored("rootdir"))
        self.assertEqual(None, tree.is_ignored("foo/rootdir"))
        self.assertEqual(None, tree.is_ignored("rootdirtrailer"))

        self.assertEqual("randomfile*", tree.is_ignored("randomfile"))
        self.assertEqual("randomfile*", tree.is_ignored("randomfiles"))
        self.assertEqual("randomfile*", tree.is_ignored("foo/randomfiles"))
        self.assertEqual(None, tree.is_ignored("randomfil"))
        self.assertEqual(None, tree.is_ignored("foo/randomfil"))

        self.assertEqual("path/from/ro?t", tree.is_ignored("path/from/root"))
        self.assertEqual("path/from/ro?t", tree.is_ignored("path/from/roat"))
        self.assertEqual(None, tree.is_ignored("roat"))

        self.assertEqual("**/piffle.py", tree.is_ignored("piffle.py"))
        self.assertEqual("**/piffle.py", tree.is_ignored("a/piffle.py"))
        self.assertEqual(None, tree.is_ignored("b/piffle.py"))  # exclusion
        self.assertEqual("**/piffle.py", tree.is_ignored("foo/bar/piffle.py"))
        self.assertEqual(None, tree.is_ignored("p/iffle.py"))

        self.assertEqual("unicode\xb5", tree.is_ignored("unicode\xb5"))
        self.assertEqual("unicode\xb5", tree.is_ignored("subdir/unicode\xb5"))
        self.assertEqual(None, tree.is_ignored("unicode\xe5"))
        self.assertEqual(None, tree.is_ignored("unicode"))
        self.assertEqual(None, tree.is_ignored("\xb5"))

        self.assertEqual("dos", tree.is_ignored("dos"))
        self.assertEqual(None, tree.is_ignored("dosfoo"))

        self.assertEqual("*bar", tree.is_ignored("foobar"))
        self.assertEqual("*bar", tree.is_ignored(r"foo\nbar"))
        self.assertEqual("*bar", tree.is_ignored("bar"))
        self.assertEqual("*bar", tree.is_ignored(".bar"))

        self.assertEqual(None, tree.is_ignored("bazbar"))  # exclusion

        self.assertEqual("?foo", tree.is_ignored("afoo"))
        self.assertEqual("?foo", tree.is_ignored(".foo"))

        self.assertEqual("*.~*", tree.is_ignored("blah.py.~1~"))

        self.assertEqual("dir1/*f1", tree.is_ignored("dir1/foof1"))
        self.assertEqual("dir1/*f1", tree.is_ignored("dir1/f1"))
        self.assertEqual("dir1/*f1", tree.is_ignored("dir1/.f1"))

        self.assertEqual("dir1/?f2", tree.is_ignored("dir1/ff2"))
        self.assertEqual("dir1/?f2", tree.is_ignored("dir1/.f2"))

        self.assertEqual("RE:dir2/.*\\.wombat", tree.is_ignored("dir2/foo.wombat"))
        self.assertEqual(None, tree.is_ignored("dir2/foo"))

        # Blank lines and comments should be ignored
        self.assertEqual(None, tree.is_ignored(""))
        self.assertEqual(None, tree.is_ignored("test/"))

        self.assertEqual(None, tree.is_ignored("#comment"))

        # Whitespace should not be stripped
        self.assertEqual(" xx ", tree.is_ignored(" xx "))
        self.assertEqual(" xx ", tree.is_ignored("subdir/ xx "))
        self.assertEqual(None, tree.is_ignored("xx"))
        self.assertEqual(None, tree.is_ignored("xx "))
        self.assertEqual(None, tree.is_ignored(" xx"))
        self.assertEqual(None, tree.is_ignored("subdir/xx "))

    def test_global_ignored(self):
        tree = self.make_branch_and_tree(".")

        bedding.ensure_config_dir_exists()
        bedding.user_ignore_config_path()
        self._set_user_ignore_content(
            b"*.py[co]\n"
            b"./.shelf\n"
            b"# comment line\n"
            b"\n"  # Blank line
            b"\r\n"  # Blank dos line
            b" * \n"  # Trailing and suffix spaces
            b"crlf\r\n"  # dos style line
            b"*\xc3\xa5*\n"  # u'\xe5'.encode('utf8')
        )

        # Rooted
        self.assertEqual("./.shelf", tree.is_ignored(".shelf"))
        self.assertEqual(None, tree.is_ignored("foo/.shelf"))

        # Glob style
        self.assertEqual("*.py[co]", tree.is_ignored("foo.pyc"))
        self.assertEqual("*.py[co]", tree.is_ignored("foo.pyo"))
        self.assertEqual(None, tree.is_ignored("foo.py"))

        # Glob in subdir
        self.assertEqual("*.py[co]", tree.is_ignored("bar/foo.pyc"))
        self.assertEqual("*.py[co]", tree.is_ignored("bar/foo.pyo"))
        self.assertEqual(None, tree.is_ignored("bar/foo.py"))

        # Unicode
        self.assertEqual("*\xe5*", tree.is_ignored("b\xe5gfors"))
        self.assertEqual("*\xe5*", tree.is_ignored("\xe5gfors"))
        self.assertEqual("*\xe5*", tree.is_ignored("\xe5"))
        self.assertEqual("*\xe5*", tree.is_ignored("b\xe5"))
        self.assertEqual("*\xe5*", tree.is_ignored("b/\xe5"))

        # Whitespace
        self.assertEqual(" * ", tree.is_ignored(" bbb "))
        self.assertEqual(" * ", tree.is_ignored("subdir/ bbb "))
        self.assertEqual(None, tree.is_ignored("bbb "))
        self.assertEqual(None, tree.is_ignored(" bbb"))

        # Dos lines
        self.assertEqual("crlf", tree.is_ignored("crlf"))
        self.assertEqual("crlf", tree.is_ignored("subdir/crlf"))

        # Comment line should be ignored
        self.assertEqual(None, tree.is_ignored("# comment line"))

        # Blank line should also be ignored
        self.assertEqual(None, tree.is_ignored(""))
        self.assertEqual(None, tree.is_ignored("baz/"))

    def test_mixed_is_ignored(self):
        tree = self.make_branch_and_tree(".")
        ignores._set_user_ignores(["*.py[co]", "./.shelf"])
        self.build_tree_contents([(".bzrignore", b"./rootdir\n*.swp\n")])

        self.assertEqual("*.py[co]", tree.is_ignored("foo.pyc"))
        self.assertEqual("./.shelf", tree.is_ignored(".shelf"))
        self.assertEqual("./rootdir", tree.is_ignored("rootdir"))
        self.assertEqual("*.swp", tree.is_ignored("foo.py.swp"))
        self.assertEqual("*.swp", tree.is_ignored(".foo.py.swp"))
        self.assertEqual(None, tree.is_ignored(".foo.py.swo"))

    def test_runtime_ignores(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree_contents([(".bzrignore", b"")])
        ignores._set_user_ignores([])

        orig_runtime = ignores._runtime_ignores
        try:
            ignores._runtime_ignores = set()
            self.assertEqual(None, tree.is_ignored("foobar.py"))

            tree._flush_ignore_list_cache()
            ignores.add_runtime_ignores(["./foobar.py"])
            self.assertEqual({"./foobar.py"}, ignores.get_runtime_ignores())
            self.assertEqual("./foobar.py", tree.is_ignored("foobar.py"))
        finally:
            ignores._runtime_ignores = orig_runtime

    def test_ignore_caching(self):
        tree = self.make_branch_and_tree(".")
        self.build_tree(["ignoreme"])

        self.assertEqual(None, tree.is_ignored("ignoreme"))

        # Bug #129694 specifically references WorkingTree.unknowns()
        tree.unknowns()

        self.build_tree_contents([(".bzrignore", b"ignoreme")])
        self.assertEqual("ignoreme", tree.is_ignored("ignoreme"))



# Copyright (C) 2010, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests of breezy test matchers."""

from testtools.matchers import *

from . import TestCase, TestCaseWithTransport
from .matchers import *


class StubTree:
    """Stubg for testing."""

    def __init__(self, lock_status):
        self._is_locked = lock_status

    def __str__(self):
        return "I am da tree"

    def is_locked(self):
        return self._is_locked


class FakeUnlockable:
    """Something that can be unlocked."""

    def unlock(self):
        pass


class TestReturnsUnlockable(TestCase):
    def test___str__(self):
        matcher = ReturnsUnlockable(StubTree(True))
        self.assertEqual("ReturnsUnlockable(lockable_thing=I am da tree)", str(matcher))

    def test_match(self):
        stub_tree = StubTree(False)
        matcher = ReturnsUnlockable(stub_tree)
        self.assertThat(matcher.match(lambda: FakeUnlockable()), Equals(None))

    def test_mismatch(self):
        stub_tree = StubTree(True)
        matcher = ReturnsUnlockable(stub_tree)
        mismatch = matcher.match(lambda: FakeUnlockable())
        self.assertNotEqual(None, mismatch)
        self.assertThat(mismatch.describe(), Equals("I am da tree is locked"))


class TestMatchesAncestry(TestCaseWithTransport):
    def test__str__(self):
        matcher = MatchesAncestry("A repository", b"arevid")
        self.assertEqual(
            "MatchesAncestry(repository='A repository', revision_id={!r})".format(
                b"arevid"
            ),
            str(matcher),
        )

    def test_match(self):
        b = self.make_branch_builder(".")
        b.start_series()
        revid1 = b.build_commit()
        revid2 = b.build_commit()
        b.finish_series()
        branch = b.get_branch()
        m = MatchesAncestry(branch.repository, revid2)
        self.assertThat([revid2, revid1], m)
        self.assertThat([revid1, revid2], m)
        m = MatchesAncestry(branch.repository, revid1)
        self.assertThat([revid1], m)
        m = MatchesAncestry(branch.repository, b"unknown")
        self.assertThat([b"unknown"], m)

    def test_mismatch(self):
        b = self.make_branch_builder(".")
        b.start_series()
        revid1 = b.build_commit()
        b.build_commit()
        b.finish_series()
        branch = b.get_branch()
        m = MatchesAncestry(branch.repository, revid1)
        mismatch = m.match([])
        self.assertIsNot(None, mismatch)
        self.assertEqual(
            "mismatched ancestry for revision {!r} was [{!r}], expected []".format(
                revid1, revid1
            ),
            mismatch.describe(),
        )


class TestHasLayout(TestCaseWithTransport):
    def test__str__(self):
        matcher = HasLayout([(b"a", b"a-id")])
        self.assertEqual("HasLayout({!r})".format([(b"a", b"a-id")]), str(matcher))

    def test_match(self):
        t = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        t.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])
        self.assertThat(t, HasLayout(["", "a", "b/", "b/c"]))
        self.assertThat(
            t,
            HasLayout(
                [("", t.path2id("")), ("a", b"a-id"), ("b/", b"b-id"), ("b/c", b"c-id")]
            ),
        )

    def test_mismatch(self):
        t = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        t.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])
        mismatch = HasLayout(["a"]).match(t)
        self.assertIsNot(None, mismatch)
        self.assertEqual(
            {"['', 'a', 'b/', 'b/c']", "['a']"}, set(mismatch.describe().split(" != "))
        )

    def test_no_dirs(self):
        # Some tree/repository formats do not support versioned directories
        t = self.make_branch_and_tree(".")
        t.has_versioned_directories = lambda: False
        self.build_tree(["a", "b/", "b/c"])
        t.add(["a", "b", "b/c"], ids=[b"a-id", b"b-id", b"c-id"])
        self.assertIs(None, HasLayout(["", "a", "b/", "b/c"]).match(t))
        self.assertIs(None, HasLayout(["", "a", "b/", "b/c", "d/"]).match(t))
        mismatch = HasLayout(["", "a", "d/"]).match(t)
        self.assertIsNot(None, mismatch)
        self.assertEqual(
            {"['', 'a', 'b/', 'b/c']", "['', 'a']"},
            set(mismatch.describe().split(" != ")),
        )


class TestHasPathRelations(TestCaseWithTransport):
    def test__str__(self):
        t = self.make_branch_and_tree(".")
        matcher = HasPathRelations(t, [("a", "b")])
        self.assertEqual(
            "HasPathRelations({!r}, {!r})".format(t, [("a", "b")]), str(matcher)
        )

    def test_match(self):
        t = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        t.add(["a", "b", "b/c"])
        self.assertThat(
            t, HasPathRelations(t, [("", ""), ("a", "a"), ("b/", "b/"), ("b/c", "b/c")])
        )

    def test_mismatch(self):
        t = self.make_branch_and_tree(".")
        self.build_tree(["a", "b/", "b/c"])
        t.add(["a", "b", "b/c"])
        mismatch = HasPathRelations(t, [("a", "a")]).match(t)
        self.assertIsNot(None, mismatch)


class TestRevisionHistoryMatches(TestCaseWithTransport):
    def test_empty(self):
        tree = self.make_branch_and_tree(".")
        matcher = RevisionHistoryMatches([])
        self.assertIs(None, matcher.match(tree.branch))

    def test_matches(self):
        tree = self.make_branch_and_tree(".")
        tree.commit("msg1", rev_id=b"a")
        tree.commit("msg2", rev_id=b"b")
        matcher = RevisionHistoryMatches([b"a", b"b"])
        self.assertIs(None, matcher.match(tree.branch))

    def test_mismatch(self):
        tree = self.make_branch_and_tree(".")
        tree.commit("msg1", rev_id=b"a")
        tree.commit("msg2", rev_id=b"b")
        matcher = RevisionHistoryMatches([b"a", b"b", b"c"])
        self.assertEqual(
            {"[b'a', b'b']", "[b'a', b'b', b'c']"},
            set(matcher.match(tree.branch).describe().split(" != ")),
        )



# Copyright (C) 2006-2010 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the WorkingTree.merge_from_branch api."""

import os

from breezy import errors, merge
from breezy.tests import per_workingtree
from breezy.workingtree import PointlessMerge


class TestMergeFromBranch(per_workingtree.TestCaseWithWorkingTree):
    def create_two_trees_for_merging(self):
        """Create two trees that can be merged from.

        This sets self.tree_from, self.first_rev, self.tree_to, self.second_rev
        and self.to_second_rev.
        """
        self.tree_from = self.make_branch_and_tree("from")
        self.first_rev = self.tree_from.commit("first post")
        self.tree_to = self.tree_from.controldir.sprout("to").open_workingtree()
        self.second_rev = self.tree_from.commit(
            "second rev on from", allow_pointless=True
        )
        self.to_second_rev = self.tree_to.commit(
            "second rev on to", allow_pointless=True
        )

    def test_smoking_merge(self):
        """Smoke test of merge_from_branch."""
        self.create_two_trees_for_merging()
        self.tree_to.merge_from_branch(self.tree_from.branch)
        self.assertEqual(
            [self.to_second_rev, self.second_rev], self.tree_to.get_parent_ids()
        )

    def test_merge_to_revision(self):
        """Merge from a branch to a revision that is not the tip."""
        self.create_two_trees_for_merging()
        self.third_rev = self.tree_from.commit("real_tip")
        self.tree_to.merge_from_branch(
            self.tree_from.branch, to_revision=self.second_rev
        )
        self.assertEqual(
            [self.to_second_rev, self.second_rev], self.tree_to.get_parent_ids()
        )

    def test_compare_after_merge(self):
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree_contents([("tree_a/file", b"text-a")])
        tree_a.add("file")
        tree_a.commit("added file")
        tree_b = tree_a.controldir.sprout("tree_b").open_workingtree()
        os.unlink("tree_a/file")
        tree_a.commit("deleted file")
        self.build_tree_contents([("tree_b/file", b"text-b")])
        tree_b.commit("changed file")
        tree_a.merge_from_branch(tree_b.branch)
        tree_a.lock_read()
        self.addCleanup(tree_a.unlock)
        list(tree_a.iter_changes(tree_a.basis_tree()))

    def test_merge_empty(self):
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree_contents([("tree_a/file", b"text-a")])
        tree_a.add("file")
        tree_a.commit("added file")
        tree_b = self.make_branch_and_tree("treeb")
        self.assertRaises(errors.NoCommits, tree_a.merge_from_branch, tree_b.branch)
        tree_b.merge_from_branch(tree_a.branch)

    def test_merge_base(self):
        tree_a = self.make_branch_and_tree("tree_a")
        self.build_tree_contents([("tree_a/file", b"text-a")])
        tree_a.add("file")
        rev1 = tree_a.commit("added file")
        tree_b = tree_a.controldir.sprout("tree_b").open_workingtree()
        os.unlink("tree_a/file")
        tree_a.commit("deleted file")
        self.build_tree_contents([("tree_b/file", b"text-b")])
        tree_b.commit("changed file")
        self.assertRaises(
            PointlessMerge,
            tree_a.merge_from_branch,
            tree_b.branch,
            from_revision=tree_b.branch.last_revision(),
        )
        tree_a.merge_from_branch(tree_b.branch, from_revision=rev1)
        tree_a.lock_read()
        self.addCleanup(tree_a.unlock)
        changes = list(tree_a.iter_changes(tree_a.basis_tree()))
        self.assertEqual(1, len(changes), changes)

    def test_merge_type(self):
        this = self.make_branch_and_tree("this")
        self.build_tree_contents([("this/foo", b"foo")])
        this.add("foo")
        this.commit("added foo")
        other = this.controldir.sprout("other").open_workingtree()
        self.build_tree_contents([("other/foo", b"bar")])
        other.commit("content -> bar")
        self.build_tree_contents([("this/foo", b"baz")])
        this.commit("content -> baz")

        class QuxMerge(merge.Merge3Merger):
            def text_merge(self, trans_id, paths):
                self.tt.create_file([b"qux"], trans_id)

        this.merge_from_branch(other.branch, merge_type=QuxMerge)
        self.assertEqual(b"qux", this.get_file_text("foo"))


class TestMergedBranch(per_workingtree.TestCaseWithWorkingTree):
    def make_inner_branch(self):
        bld_inner = self.make_branch_builder("inner")
        bld_inner.start_series()
        rev1 = bld_inner.build_snapshot(
            None,
            [
                ("add", ("", None, "directory", "")),
                ("add", ("dir", None, "directory", "")),
                ("add", ("dir/file1", None, "file", b"file1 content\n")),
                ("add", ("file3", None, "file", b"file3 content\n")),
            ],
        )
        rev4 = bld_inner.build_snapshot(
            [rev1], [("add", ("file4", None, "file", b"file4 content\n"))]
        )
        rev5 = bld_inner.build_snapshot([rev4], [("rename", ("file4", "dir/file4"))])
        rev3 = bld_inner.build_snapshot(
            [rev1],
            [
                ("modify", ("file3", b"new file3 contents\n")),
            ],
        )
        rev2 = bld_inner.build_snapshot(
            [rev1],
            [
                ("add", ("dir/file2", None, "file", b"file2 content\n")),
            ],
        )
        bld_inner.finish_series()
        br = bld_inner.get_branch()
        return br, [rev1, rev2, rev3, rev4, rev5]

    def assertTreeLayout(self, expected, tree):
        with tree.lock_read():
            actual = [e[0] for e in tree.list_files()]
            # list_files doesn't guarantee order
            actual = sorted(actual)
            self.assertEqual(expected, actual)

    def make_outer_tree(self):
        outer = self.make_branch_and_tree("outer")
        self.build_tree_contents([("outer/foo", b"foo")])
        outer.add("foo")
        outer.commit("added foo")
        inner, revs = self.make_inner_branch()
        outer.merge_from_branch(inner, to_revision=revs[0], from_revision=b"null:")
        # retain original root id.
        if outer.supports_setting_file_ids():
            outer.set_root_id(outer.basis_tree().path2id(""))
        outer.commit("merge inner branch")
        outer.mkdir("dir-outer")
        outer.move(["dir", "file3"], to_dir="dir-outer")
        outer.commit("rename imported dir and file3 to dir-outer")
        return outer, inner, revs

    def test_file1_deleted_in_dir(self):
        outer, inner, revs = self.make_outer_tree()
        outer.remove(["dir-outer/dir/file1"], keep_files=False)
        outer.commit("delete file1")
        outer.merge_from_branch(inner)
        outer.commit("merge the rest")
        if outer.supports_rename_tracking():
            self.assertTreeLayout(
                [
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/dir/file2",
                    "dir-outer/file3",
                    "foo",
                ],
                outer,
            )
        else:
            self.assertTreeLayout(
                [
                    "dir",
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/file3",
                    "dir/file2",
                    "foo",
                ],
                outer,
            )

    def test_file3_deleted_in_root(self):
        # Reproduce bug #375898
        outer, inner, revs = self.make_outer_tree()
        outer.remove(["dir-outer/file3"], keep_files=False)
        outer.commit("delete file3")
        outer.merge_from_branch(inner)
        outer.commit("merge the rest")
        if outer.supports_rename_tracking():
            self.assertTreeLayout(
                [
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/dir/file1",
                    "dir-outer/dir/file2",
                    "foo",
                ],
                outer,
            )
        else:
            self.assertTreeLayout(
                [
                    "dir",
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/dir/file1",
                    "dir/file2",
                    "foo",
                ],
                outer,
            )

    def test_file3_in_root_conflicted(self):
        outer, inner, revs = self.make_outer_tree()
        outer.remove(["dir-outer/file3"], keep_files=False)
        outer.commit("delete file3")
        nb_conflicts = outer.merge_from_branch(inner, to_revision=revs[2])
        if outer.supports_rename_tracking():
            self.assertEqual(4, len(nb_conflicts))
        else:
            self.assertEqual(1, len(nb_conflicts))
        self.assertTreeLayout(
            [
                "dir-outer",
                "dir-outer/dir",
                "dir-outer/dir/file1",
                # Ideally th conflict helpers should be in
                # dir-outer/dir but since we can't easily find
                # back the file3 -> outer-dir/dir rename, root
                # is good enough -- vila 20100401
                "file3.BASE",
                "file3.OTHER",
                "foo",
            ],
            outer,
        )

    def test_file4_added_in_root(self):
        outer, inner, revs = self.make_outer_tree()
        nb_conflicts = outer.merge_from_branch(inner, to_revision=revs[3])
        # file4 could not be added to its original root, so it gets added to
        # the new root with a conflict.
        if outer.supports_rename_tracking():
            self.assertEqual(1, len(nb_conflicts))
        else:
            self.assertEqual(0, len(nb_conflicts))
        self.assertTreeLayout(
            [
                "dir-outer",
                "dir-outer/dir",
                "dir-outer/dir/file1",
                "dir-outer/file3",
                "file4",
                "foo",
            ],
            outer,
        )

    def test_file4_added_then_renamed(self):
        outer, inner, revs = self.make_outer_tree()
        # 1 conflict, because file4 can't be put into the old root
        nb_conflicts = outer.merge_from_branch(inner, to_revision=revs[3])
        if outer.supports_rename_tracking():
            self.assertEqual(1, len(nb_conflicts))
        else:
            self.assertEqual(0, len(nb_conflicts))
        try:
            outer.set_conflicts([])
        except errors.UnsupportedOperation:
            # WT2 doesn't have a separate list of conflicts to clear. It
            # actually says there is a conflict, but happily forgets all about
            # it.
            pass
        outer.commit("added file4")
        # And now file4 gets renamed into an existing dir
        nb_conflicts = outer.merge_from_branch(inner, to_revision=revs[4])
        if outer.supports_rename_tracking():
            self.assertEqual(1, len(nb_conflicts))
            self.assertTreeLayout(
                [
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/dir/file1",
                    "dir-outer/dir/file4",
                    "dir-outer/file3",
                    "foo",
                ],
                outer,
            )
        else:
            if outer.has_versioned_directories():
                self.assertEqual(2, len(nb_conflicts))
            else:
                self.assertEqual(0, len(nb_conflicts))
            self.assertTreeLayout(
                [
                    "dir",
                    "dir-outer",
                    "dir-outer/dir",
                    "dir-outer/dir/file1",
                    "dir-outer/file3",
                    "dir/file4",
                    "foo",
                ],
                outer,
            )



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the contract of commit on branches."""

from breezy import branch, revision
from breezy.tests import per_branch


class TestCommit(per_branch.TestCaseWithBranch):
    def test_commit_nicks(self):
        """Nicknames are committed to the revision."""
        self.get_transport().mkdir("bzr.dev")
        wt = self.make_branch_and_tree("bzr.dev")
        branch = wt.branch
        branch.nick = "My happy branch"
        wt.commit("My commit respect da nick.")
        committed = branch.repository.get_revision(branch.last_revision())
        if branch.repository._format.supports_storing_branch_nick:
            self.assertEqual(committed.properties["branch-nick"], "My happy branch")
        else:
            self.assertNotIn("branch-nick", committed.properties)


class TestCommitHook(per_branch.TestCaseWithBranch):
    def setUp(self):
        self.hook_calls = []
        super().setUp()

    def capture_post_commit_hook(
        self, local, master, old_revno, old_revid, new_revno, new_revid
    ):
        """Capture post commit hook calls to self.hook_calls.

        The call is logged, as is some state of the two branches.
        """
        if local:
            local_locked = local.is_locked()
            local_base = local.base
        else:
            local_locked = None
            local_base = None
        self.hook_calls.append(
            (
                "post_commit",
                local_base,
                master.base,
                old_revno,
                old_revid,
                new_revno,
                new_revid,
                local_locked,
                master.is_locked(),
            )
        )

    def capture_pre_commit_hook(
        self,
        local,
        master,
        old_revno,
        old_revid,
        new_revno,
        new_revid,
        tree_delta,
        future_tree,
    ):
        self.hook_calls.append(
            ("pre_commit", old_revno, old_revid, new_revno, new_revid, tree_delta)
        )

    def test_post_commit_to_origin(self):
        tree = self.make_branch_and_memory_tree("branch")
        branch.Branch.hooks.install_named_hook(
            "post_commit", self.capture_post_commit_hook, None
        )
        tree.lock_write()
        tree.add("")
        revid = tree.commit("a revision")
        # should have had one notification, from origin, and
        # have the branch locked at notification time.
        self.assertEqual(
            [
                (
                    "post_commit",
                    None,
                    tree.branch.base,
                    0,
                    revision.NULL_REVISION,
                    1,
                    revid,
                    None,
                    True,
                )
            ],
            self.hook_calls,
        )
        tree.unlock()

    def test_post_commit_bound(self):
        master = self.make_branch("master")
        tree = self.make_branch_and_memory_tree("local")
        try:
            tree.branch.bind(master)
        except branch.BindingUnsupported:
            # cant bind this format, the test is irrelevant.
            return
        branch.Branch.hooks.install_named_hook(
            "post_commit", self.capture_post_commit_hook, None
        )
        tree.lock_write()
        tree.add("")
        revid = tree.commit("a revision")
        # with a bound branch, local is set.
        self.assertEqual(
            [
                (
                    "post_commit",
                    tree.branch.base,
                    master.base,
                    0,
                    revision.NULL_REVISION,
                    1,
                    revid,
                    True,
                    True,
                )
            ],
            self.hook_calls,
        )
        tree.unlock()

    def test_post_commit_not_to_origin(self):
        tree = self.make_branch_and_memory_tree("branch")
        with tree.lock_write():
            tree.add("")
            revid = tree.commit("first revision")
            branch.Branch.hooks.install_named_hook(
                "post_commit", self.capture_post_commit_hook, None
            )
            revid2 = tree.commit("second revision")
            # having committed from up the branch, we should get the
            # before and after revnos and revids correctly.
            self.assertEqual(
                [
                    (
                        "post_commit",
                        None,
                        tree.branch.base,
                        1,
                        revid,
                        2,
                        revid2,
                        None,
                        True,
                    )
                ],
                self.hook_calls,
            )

    def get_rootfull_delta(self, repository, revid):
        tree = repository.revision_tree(revid)
        with repository.lock_read():
            parent_revid = repository.get_parent_map([revid])[revid][0]
            basis_tree = repository.revision_tree(parent_revid)
            tree = repository.revision_tree(revid)
            return tree.changes_from(basis_tree, include_root=True)

    def test_pre_commit_passes(self):
        tree = self.make_branch_and_memory_tree("branch")
        with tree.lock_write():
            tree.add("")
            branch.Branch.hooks.install_named_hook(
                "pre_commit", self.capture_pre_commit_hook, None
            )
            revid1 = tree.commit("first revision")
            revid2 = tree.commit("second revision")
            root_delta = self.get_rootfull_delta(tree.branch.repository, revid1)
            empty_delta = tree.branch.repository.get_revision_delta(revid2)
            self.assertEqual(
                [
                    ("pre_commit", 0, revision.NULL_REVISION, 1, revid1, root_delta),
                    ("pre_commit", 1, revid1, 2, revid2, empty_delta),
                ],
                self.hook_calls,
            )

    def test_pre_commit_fails(self):
        tree = self.make_branch_and_memory_tree("branch")
        with tree.lock_write():
            tree.add("")

            class PreCommitException(Exception):
                def __init__(self, revid):
                    self.revid = revid

            def hook_func(
                local,
                master,
                old_revno,
                old_revid,
                new_revno,
                new_revid,
                tree_delta,
                future_tree,
            ):
                raise PreCommitException(new_revid)

            branch.Branch.hooks.install_named_hook(
                "pre_commit", self.capture_pre_commit_hook, None
            )
            branch.Branch.hooks.install_named_hook("pre_commit", hook_func, None)
            revids = [None, None, None]
            # this commit will raise an exception
            # so the commit is rolled back and revno unchanged
            err = self.assertRaises(PreCommitException, tree.commit, "message")
            # we have to record the revid to use in assertEqual later
            revids[0] = err.revid
            # unregister all pre_commit hooks
            branch.Branch.hooks["pre_commit"] = []
            # and re-register the capture hook
            branch.Branch.hooks.install_named_hook(
                "pre_commit", self.capture_pre_commit_hook, None
            )
            # now these commits should go through
            for i in range(1, 3):
                revids[i] = tree.commit("message")
            self.assertEqual(
                [
                    (
                        "pre_commit",
                        0,
                        revision.NULL_REVISION,
                        1,
                        revids[0],
                        self.get_rootfull_delta(tree.branch.repository, revids[0]),
                    ),
                    (
                        "pre_commit",
                        0,
                        revision.NULL_REVISION,
                        1,
                        revids[1],
                        self.get_rootfull_delta(tree.branch.repository, revids[1]),
                    ),
                    (
                        "pre_commit",
                        1,
                        revids[1],
                        2,
                        revids[2],
                        self.get_rootfull_delta(tree.branch.repository, revids[2]),
                    ),
                ],
                self.hook_calls,
            )

    def test_pre_commit_delta(self):
        # This tests the TreeDelta object passed to pre_commit hook.
        # This does not try to validate data correctness in the delta.
        self.build_tree(["rootfile", "dir/", "dir/subfile"])
        tree = self.make_branch_and_tree(".")
        with tree.lock_write():
            # setting up a playground
            tree.add("rootfile")
            tree.put_file_bytes_non_atomic("rootfile", b"abc")
            tree.add("dir")
            tree.add("dir/subfile")
            tree.put_file_bytes_non_atomic("to_be_unversioned", b"blah")
            tree.add(["to_be_unversioned"])
            tree.put_file_bytes_non_atomic("dir/subfile", b"def")
            revid1 = tree.commit("first revision")

        with tree.lock_write():
            # making changes
            tree.put_file_bytes_non_atomic("rootfile", b"jkl")
            tree.rename_one("dir/subfile", "dir/subfile_renamed")
            tree.unversion(["to_be_unversioned"])
            tree.mkdir("added_dir")
            # start to capture pre_commit delta
            branch.Branch.hooks.install_named_hook(
                "pre_commit", self.capture_pre_commit_hook, None
            )
            revid2 = tree.commit("second revision")

        self.assertEqual(
            [
                (
                    "pre_commit",
                    1,
                    revid1,
                    2,
                    revid2,
                    self.get_rootfull_delta(tree.branch.repository, revid2),
                )
            ],
            self.hook_calls,
        )



# Copyright (C) 2006-2012, 2016 Canonical Ltd
# Authors: Robert Collins <robert.collins@canonical.com>
#          and others
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Repository implementation tests for bzr.

These test the conformance of all the repository variations to the expected API.
Specific tests for individual formats are in the tests/test_repository.py file
rather than in tests/per_branch/*.py.
"""

from breezy import repository
from breezy.bzr.remote import RemoteRepositoryFormat
from breezy.tests import default_transport, multiply_tests, test_server
from breezy.tests.per_controldir.test_controldir import TestCaseWithControlDir
from breezy.transport import memory


def formats_to_scenarios(
    formats, transport_server, transport_readonly_server, vfs_transport_factory=None
):
    """Transform the input formats to a list of scenarios.

    :param formats: A list of (scenario_name_suffix, repo_format)
        where the scenario_name_suffix is to be appended to the format
        name, and the repo_format is a RepositoryFormat subclass
        instance.
    :returns: Scenarios of [(scenario_name, {parameter_name: value})]
    """
    result = []
    for scenario_name_suffix, repository_format in formats:
        scenario_name = repository_format.__class__.__name__
        scenario_name += scenario_name_suffix
        scenario = (
            scenario_name,
            {
                "transport_server": transport_server,
                "transport_readonly_server": transport_readonly_server,
                "bzrdir_format": repository_format._matchingcontroldir,
                "repository_format": repository_format,
            },
        )
        # Only override the test's vfs_transport_factory if one was
        # specified, otherwise just leave the default in place.
        if vfs_transport_factory:
            scenario[1]["vfs_transport_factory"] = vfs_transport_factory
        result.append(scenario)
    return result


def all_repository_format_scenarios():
    """Return a list of test scenarios for parameterising repository tests."""
    all_formats = repository.format_registry._get_all()
    # format_scenarios is all the implementations of Repository; i.e. all disk
    # formats plus RemoteRepository.
    format_scenarios = formats_to_scenarios(
        [("", format) for format in all_formats],
        default_transport,
        # None here will cause a readonly decorator to be created
        # by the TestCaseWithTransport.get_readonly_transport method.
        None,
    )
    format_scenarios.extend(
        formats_to_scenarios(
            [("-default", RemoteRepositoryFormat())],
            test_server.SmartTCPServer_for_testing,
            test_server.ReadonlySmartTCPServer_for_testing,
            memory.MemoryServer,
        )
    )
    format_scenarios.extend(
        formats_to_scenarios(
            [("-v2", RemoteRepositoryFormat())],
            test_server.SmartTCPServer_for_testing_v2_only,
            test_server.ReadonlySmartTCPServer_for_testing_v2_only,
            memory.MemoryServer,
        )
    )
    return format_scenarios


class TestCaseWithRepository(TestCaseWithControlDir):
    def get_default_format(self):
        format = self.repository_format._matchingcontroldir
        self.assertEqual(format.repository_format, self.repository_format)
        return format

    def make_repository(self, relpath, shared=None, format=None):
        format = self.resolve_format(format)
        repo = super().make_repository(relpath, shared=shared, format=format)
        if format is None or format.repository_format is self.repository_format:
            # Create a repository of the type we are trying to test.
            if getattr(self, "repository_to_test_repository", None):
                repo = self.repository_to_test_repository(repo)
        return repo


def load_tests(loader, standard_tests, pattern):
    prefix = "breezy.tests.per_repository."
    test_repository_modules = [
        "test_add_fallback_repository",
        "test_break_lock",
        "test_check",
        "test_commit_builder",
        "test_fetch",
        "test_file_graph",
        "test_get_parent_map",
        "test_get_rev_id_for_revno",
        "test_has_same_location",
        "test_has_revisions",
        "test_locking",
        "test_pack",
        "test_reconcile",
        "test_refresh_data",
        "test_repository",
        "test_revision",
        "test_signatures",
        "test_statistics",
        "test_write_group",
    ]
    # Parameterize per_repository test modules by format.
    submod_tests = loader.loadTestsFromModuleNames(
        [prefix + module_name for module_name in test_repository_modules]
    )
    format_scenarios = all_repository_format_scenarios()
    return multiply_tests(submod_tests, format_scenarios, standard_tests)



# Copyright (C) 2022 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for GitHub."""

from datetime import datetime

from ....tests import TestCase
from ..forge import parse_timestring


class ParseTimestringTests(TestCase):
    def test_simple(self):
        self.assertEqual(
            datetime(2011, 1, 26, 19, 1, 12), parse_timestring("2011-01-26T19:01:12Z")
        )



# Copyright (C) 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for breezy.bzr.inventory_delta.

See doc/developer/inventory.txt for more information.
"""

from io import BytesIO

from ... import osutils
from ...revision import NULL_REVISION
from .. import inventory, inventory_delta
from ..inventory import Inventory
from ..inventory_delta import InventoryDeltaError
from . import TestCase

### DO NOT REFLOW THESE TEXTS. NEW LINES ARE SIGNIFICANT. ###
empty_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: null:
versioned_root: true
tree_references: true
"""

root_only_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: entry-version
versioned_root: true
tree_references: true
None\x00/\x00an-id\x00\x00a@e\xc3\xa5ample.com--2004\x00dir
"""


root_change_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: entry-version
version: changed-root
versioned_root: true
tree_references: true
/\x00an-id\x00\x00different-version\x00dir
"""

corrupt_parent_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: entry-version
version: changed-root
versioned_root: false
tree_references: false
/\x00an-id\x00\x00different-version\x00dir
"""

root_only_unversioned = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: entry-version
versioned_root: false
tree_references: false
None\x00/\x00TREE_ROOT\x00\x00entry-version\x00dir
"""

reference_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: entry-version
versioned_root: true
tree_references: true
None\x00/\x00TREE_ROOT\x00\x00a@e\xc3\xa5ample.com--2004\x00dir
None\x00/foo\x00id\x00TREE_ROOT\x00changed\x00tree\x00subtree-version
"""

change_tree_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: entry-version
version: change-tree
versioned_root: false
tree_references: false
/foo\x00id\x00TREE_ROOT\x00changed-twice\x00tree\x00subtree-version2
"""


class TestDeserialization(TestCase):
    """Test InventoryDeltaSerializer.parse_text_bytes."""

    def test_parse_no_bytes(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(InventoryDeltaError, deserializer.parse_text_bytes, [])
        self.assertContainsRe(str(err), "inventory delta is empty")

    def test_parse_bad_format(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError, deserializer.parse_text_bytes, [b"format: foo\n"]
        )
        self.assertContainsRe(str(err), "unknown format")

    def test_parse_no_parent(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            [b"format: bzr inventory delta v1 (bzr 1.14)\n"],
        )
        self.assertContainsRe(str(err), "missing parent: marker")

    def test_parse_no_version(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            [b"format: bzr inventory delta v1 (bzr 1.14)\n", b"parent: null:\n"],
        )
        self.assertContainsRe(str(err), "missing version: marker")

    def test_parse_duplicate_key_errors(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        double_root_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: null:
versioned_root: true
tree_references: true
None\x00/\x00an-id\x00\x00a@e\xc3\xa5ample.com--2004\x00dir\x00\x00
None\x00/\x00an-id\x00\x00a@e\xc3\xa5ample.com--2004\x00dir\x00\x00
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(double_root_lines),
        )
        self.assertContainsRe(str(err), "duplicate file id")

    def test_parse_versioned_root_only(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        parse_result = deserializer.parse_text_bytes(
            osutils.split_lines(root_only_lines)
        )
        expected_entry = inventory.make_entry("directory", "", None, b"an-id")
        expected_entry.revision = b"a@e\xc3\xa5ample.com--2004"
        self.assertEqual(
            (
                b"null:",
                b"entry-version",
                True,
                True,
                [(None, "", b"an-id", expected_entry)],
            ),
            parse_result,
        )

    def test_parse_special_revid_not_valid_last_mod(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        root_only_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: null:
versioned_root: false
tree_references: true
None\x00/\x00TREE_ROOT\x00\x00null:\x00dir\x00\x00
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(root_only_lines),
        )
        self.assertContainsRe(str(err), "special revisionid found")

    def test_parse_versioned_root_versioned_disabled(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        root_only_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: null:
versioned_root: false
tree_references: true
None\x00/\x00TREE_ROOT\x00\x00a@e\xc3\xa5ample.com--2004\x00dir\x00\x00
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(root_only_lines),
        )
        self.assertContainsRe(str(err), "Versioned root found")

    def test_parse_unique_root_id_root_versioned_disabled(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        root_only_lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: parent-id
version: a@e\xc3\xa5ample.com--2004
versioned_root: false
tree_references: true
None\x00/\x00an-id\x00\x00parent-id\x00dir\x00\x00
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(root_only_lines),
        )
        self.assertContainsRe(str(err), "Versioned root found")

    def test_parse_unversioned_root_versioning_enabled(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        parse_result = deserializer.parse_text_bytes(
            osutils.split_lines(root_only_unversioned)
        )
        expected_entry = inventory.make_entry("directory", "", None, b"TREE_ROOT")
        expected_entry.revision = b"entry-version"
        self.assertEqual(
            (
                b"null:",
                b"entry-version",
                False,
                False,
                [(None, "", b"TREE_ROOT", expected_entry)],
            ),
            parse_result,
        )

    def test_parse_versioned_root_when_disabled(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer(
            allow_versioned_root=False
        )
        err = self.assertRaises(
            inventory_delta.IncompatibleInventoryDelta,
            deserializer.parse_text_bytes,
            osutils.split_lines(root_only_lines),
        )
        self.assertEqual("versioned_root not allowed", str(err))

    def test_parse_tree_when_disabled(self):
        deserializer = inventory_delta.InventoryDeltaDeserializer(
            allow_tree_references=False
        )
        err = self.assertRaises(
            inventory_delta.IncompatibleInventoryDelta,
            deserializer.parse_text_bytes,
            osutils.split_lines(reference_lines),
        )
        self.assertEqual("Tree reference not allowed", str(err))

    def test_parse_tree_when_header_disallows(self):
        # A deserializer that allows tree_references to be set or unset.
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        # A serialised inventory delta with a header saying no tree refs, but
        # that has a tree ref in its content.
        lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: entry-version
versioned_root: false
tree_references: false
None\x00/foo\x00id\x00TREE_ROOT\x00changed\x00tree\x00subtree-version
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(lines),
        )
        self.assertContainsRe(str(err), "Tree reference found")

    def test_parse_versioned_root_when_header_disallows(self):
        # A deserializer that allows tree_references to be set or unset.
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        # A serialised inventory delta with a header saying no tree refs, but
        # that has a tree ref in its content.
        lines = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: entry-version
versioned_root: false
tree_references: false
None\x00/\x00TREE_ROOT\x00\x00a@e\xc3\xa5ample.com--2004\x00dir
"""
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(lines),
        )
        self.assertContainsRe(str(err), "Versioned root found")

    def test_parse_last_line_not_empty(self):
        """Newpath must start with / if it is not None."""
        # Trim the trailing newline from a valid serialization
        lines = root_only_lines[:-1]
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(lines),
        )
        self.assertContainsRe(str(err), "last line not empty")

    def test_parse_invalid_newpath(self):
        """Newpath must start with / if it is not None."""
        lines = empty_lines
        lines += b"None\x00bad\x00TREE_ROOT\x00\x00version\x00dir\n"
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(lines),
        )
        self.assertContainsRe(str(err), "newpath invalid")

    def test_parse_invalid_oldpath(self):
        """Oldpath must start with / if it is not None."""
        lines = root_only_lines
        lines += b"bad\x00/new\x00file-id\x00\x00version\x00dir\n"
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        err = self.assertRaises(
            InventoryDeltaError,
            deserializer.parse_text_bytes,
            osutils.split_lines(lines),
        )
        self.assertContainsRe(str(err), "oldpath invalid")

    def test_parse_new_file(self):
        """A new file is parsed correctly."""
        lines = root_only_lines
        fake_sha = b"deadbeef" * 5
        lines += (
            b"None\x00/new\x00file-id\x00an-id\x00version\x00file\x00123\x00"
            + b"\x00"
            + fake_sha
            + b"\n"
        )
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        parse_result = deserializer.parse_text_bytes(osutils.split_lines(lines))
        expected_entry = inventory.make_entry("file", "new", b"an-id", b"file-id")
        expected_entry.revision = b"version"
        expected_entry.text_size = 123
        expected_entry.text_sha1 = fake_sha
        delta = parse_result[4]
        self.assertEqual((None, "new", b"file-id", expected_entry), delta[-1])

    def test_parse_delete(self):
        lines = root_only_lines
        lines += b"/old-file\x00None\x00deleted-id\x00\x00null:\x00deleted\x00\x00\n"
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        parse_result = deserializer.parse_text_bytes(osutils.split_lines(lines))
        delta = parse_result[4]
        self.assertEqual(("old-file", None, b"deleted-id", None), delta[-1])


class TestSerialization(TestCase):
    """Tests for InventoryDeltaSerializer.delta_to_lines."""

    def test_empty_delta_to_lines(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        self.assertEqual(
            BytesIO(empty_lines).readlines(),
            serializer.delta_to_lines(NULL_REVISION, NULL_REVISION, delta),
        )

    def test_root_only_to_lines(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"an-id")
        root.revision = b"a@e\xc3\xa5ample.com--2004"
        new_inv.add(root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        self.assertEqual(
            BytesIO(root_only_lines).readlines(),
            serializer.delta_to_lines(NULL_REVISION, b"entry-version", delta),
        )

    def test_unversioned_root(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        # Implicit roots are considered modified in every revision.
        root.revision = b"entry-version"
        new_inv.add(root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=False, tree_references=False
        )
        serialized_lines = serializer.delta_to_lines(
            NULL_REVISION, b"entry-version", delta
        )
        self.assertEqual(BytesIO(root_only_unversioned).readlines(), serialized_lines)
        deserializer = inventory_delta.InventoryDeltaDeserializer()
        self.assertEqual(
            (NULL_REVISION, b"entry-version", False, False, delta),
            deserializer.parse_text_bytes(serialized_lines),
        )

    def test_unversioned_non_root_errors(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        root.revision = b"a@e\xc3\xa5ample.com--2004"
        new_inv.add(root)
        non_root = new_inv.make_entry("directory", "foo", root.file_id, b"id")
        new_inv.add(non_root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        err = self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            NULL_REVISION,
            b"entry-version",
            delta,
        )
        self.assertContainsRe(str(err), "^no version for fileid b?'id'$")

    def test_richroot_unversioned_root_errors(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        new_inv.add(root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        err = self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            NULL_REVISION,
            b"entry-version",
            delta,
        )
        self.assertContainsRe(str(err), "no version for fileid b?'TREE_ROOT'$")

    def test_nonrichroot_versioned_root_errors(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        root.revision = b"a@e\xc3\xa5ample.com--2004"
        new_inv.add(root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=False, tree_references=True
        )
        err = self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            NULL_REVISION,
            b"entry-version",
            delta,
        )
        self.assertContainsRe(str(err), "^Version present for / in b?'TREE_ROOT'")

    def test_unknown_kind_errors(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"my-rich-root-id")
        root.revision = b"changed"
        new_inv.add(root)

        class StrangeInventoryEntry(inventory.InventoryEntry):
            kind = "strange"

        non_root = StrangeInventoryEntry(b"id", "foo", root.file_id)
        non_root.revision = b"changed"
        new_inv.add(non_root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        # we expect keyerror because there is little value wrapping this.
        # This test aims to prove that it errors more than how it errors.
        err = self.assertRaises(
            KeyError, serializer.delta_to_lines, NULL_REVISION, b"entry-version", delta
        )
        self.assertEqual(("strange",), err.args)

    def test_tree_reference_disabled(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        root.revision = b"a@e\xc3\xa5ample.com--2004"
        new_inv.add(root)
        non_root = new_inv.make_entry("tree-reference", "foo", root.file_id, b"id")
        non_root.revision = b"changed"
        non_root.reference_revision = b"subtree-version"
        new_inv.add(non_root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=False
        )
        # we expect keyerror because there is little value wrapping this.
        # This test aims to prove that it errors more than how it errors.
        err = self.assertRaises(
            KeyError, serializer.delta_to_lines, NULL_REVISION, b"entry-version", delta
        )
        self.assertEqual(("tree-reference",), err.args)

    def test_tree_reference_enabled(self):
        old_inv = Inventory(None)
        new_inv = Inventory(None)
        root = new_inv.make_entry("directory", "", None, b"TREE_ROOT")
        root.revision = b"a@e\xc3\xa5ample.com--2004"
        new_inv.add(root)
        non_root = new_inv.make_entry("tree-reference", "foo", root.file_id, b"id")
        non_root.revision = b"changed"
        non_root.reference_revision = b"subtree-version"
        new_inv.add(non_root)
        delta = new_inv._make_delta(old_inv)
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        self.assertEqual(
            BytesIO(reference_lines).readlines(),
            serializer.delta_to_lines(NULL_REVISION, b"entry-version", delta),
        )

    def test_to_inventory_root_id_versioned_not_permitted(self):
        root_entry = inventory.make_entry("directory", "", None, b"TREE_ROOT")
        root_entry.revision = b"some-version"
        delta = [(None, "", b"TREE_ROOT", root_entry)]
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=False, tree_references=True
        )
        self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            b"old-version",
            b"new-version",
            delta,
        )

    def test_to_inventory_root_id_not_versioned(self):
        delta = [
            (None, "", b"an-id", inventory.make_entry("directory", "", None, b"an-id"))
        ]
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            b"old-version",
            b"new-version",
            delta,
        )

    def test_to_inventory_has_tree_not_meant_to(self):
        make_entry = inventory.make_entry
        tree_ref = make_entry("tree-reference", "foo", b"changed-in", b"ref-id")
        tree_ref.reference_revision = b"ref-revision"
        delta = [
            (None, "", b"an-id", make_entry("directory", "", b"changed-in", b"an-id")),
            (None, "foo", b"ref-id", tree_ref),
            # a file that followed the root move
        ]
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        self.assertRaises(
            InventoryDeltaError,
            serializer.delta_to_lines,
            b"old-version",
            b"new-version",
            delta,
        )

    def test_to_inventory_torture(self):
        def make_entry(kind, name, parent_id, file_id, **attrs):
            entry = inventory.make_entry(kind, name, parent_id, file_id)
            for name, value in attrs.items():
                setattr(entry, name, value)
            return entry

        # this delta is crafted to have all the following:
        # - deletes
        # - renamed roots
        # - deep dirs
        # - files moved after parent dir was renamed
        # - files with and without exec bit
        delta = [
            # new root:
            (
                None,
                "",
                b"new-root-id",
                make_entry(
                    "directory", "", None, b"new-root-id", revision=b"changed-in"
                ),
            ),
            # an old root:
            (
                "",
                "old-root",
                b"TREE_ROOT",
                make_entry(
                    "directory",
                    "subdir-now",
                    b"new-root-id",
                    b"TREE_ROOT",
                    revision=b"moved-root",
                ),
            ),
            # a file that followed the root move
            (
                "under-old-root",
                "old-root/under-old-root",
                b"moved-id",
                make_entry(
                    "file",
                    "under-old-root",
                    b"TREE_ROOT",
                    b"moved-id",
                    revision=b"old-rev",
                    executable=False,
                    text_size=30,
                    text_sha1=b"some-sha",
                ),
            ),
            # a deleted path
            ("old-file", None, b"deleted-id", None),
            # a tree reference moved to the new root
            (
                "ref",
                "ref",
                b"ref-id",
                make_entry(
                    "tree-reference",
                    "ref",
                    b"new-root-id",
                    b"ref-id",
                    reference_revision=b"tree-reference-id",
                    revision=b"new-rev",
                ),
            ),
            # a symlink now in a deep dir
            (
                "dir/link",
                "old-root/dir/link",
                b"link-id",
                make_entry(
                    "symlink",
                    "link",
                    b"deep-id",
                    b"link-id",
                    symlink_target="target",
                    revision=b"new-rev",
                ),
            ),
            # a deep dir
            (
                "dir",
                "old-root/dir",
                b"deep-id",
                make_entry(
                    "directory", "dir", b"TREE_ROOT", b"deep-id", revision=b"new-rev"
                ),
            ),
            # a file with an exec bit set
            (
                None,
                "configure",
                b"exec-id",
                make_entry(
                    "file",
                    "configure",
                    b"new-root-id",
                    b"exec-id",
                    executable=True,
                    text_size=30,
                    text_sha1=b"some-sha",
                    revision=b"old-rev",
                ),
            ),
        ]
        serializer = inventory_delta.InventoryDeltaSerializer(
            versioned_root=True, tree_references=True
        )
        lines = serializer.delta_to_lines(NULL_REVISION, b"something", delta)
        expected = b"""format: bzr inventory delta v1 (bzr 1.14)
parent: null:
version: something
versioned_root: true
tree_references: true
/\x00/old-root\x00TREE_ROOT\x00new-root-id\x00moved-root\x00dir
/dir\x00/old-root/dir\x00deep-id\x00TREE_ROOT\x00new-rev\x00dir
/dir/link\x00/old-root/dir/link\x00link-id\x00deep-id\x00new-rev\x00link\x00target
/old-file\x00None\x00deleted-id\x00\x00null:\x00deleted\x00\x00
/ref\x00/ref\x00ref-id\x00new-root-id\x00new-rev\x00tree\x00tree-reference-id
/under-old-root\x00/old-root/under-old-root\x00moved-id\x00TREE_ROOT\x00old-rev\x00file\x0030\x00\x00some-sha
None\x00/\x00new-root-id\x00\x00changed-in\x00dir
None\x00/configure\x00exec-id\x00new-root-id\x00old-rev\x00file\x0030\x00Y\x00some-sha
"""
        serialized = b"".join(lines)
        self.assertIsInstance(serialized, bytes)
        self.assertEqual(expected, serialized)


class TestContent(TestCase):
    """Test serialization of the content part of a line."""

    def test_dir(self):
        entry = inventory.make_entry("directory", "a dir", None)
        self.assertEqual(b"dir", inventory_delta._directory_content(entry))

    def test_file_0_short_sha(self):
        file_entry = inventory.make_entry("file", "a file", None, b"file-id")
        file_entry.text_sha1 = b""
        file_entry.text_size = 0
        self.assertEqual(
            b"file\x000\x00\x00", inventory_delta._file_content(file_entry)
        )

    def test_file_10_foo(self):
        file_entry = inventory.make_entry("file", "a file", None, b"file-id")
        file_entry.text_sha1 = b"foo"
        file_entry.text_size = 10
        self.assertEqual(
            b"file\x0010\x00\x00foo", inventory_delta._file_content(file_entry)
        )

    def test_file_executable(self):
        file_entry = inventory.make_entry("file", "a file", None, b"file-id")
        file_entry.executable = True
        file_entry.text_sha1 = b"foo"
        file_entry.text_size = 10
        self.assertEqual(
            b"file\x0010\x00Y\x00foo", inventory_delta._file_content(file_entry)
        )

    def test_file_without_size(self):
        file_entry = inventory.make_entry("file", "a file", None, b"file-id")
        file_entry.text_sha1 = b"foo"
        self.assertRaises(
            InventoryDeltaError, inventory_delta._file_content, file_entry
        )

    def test_file_without_sha1(self):
        file_entry = inventory.make_entry("file", "a file", None, b"file-id")
        file_entry.text_size = 10
        self.assertRaises(
            InventoryDeltaError, inventory_delta._file_content, file_entry
        )

    def test_link_empty_target(self):
        entry = inventory.make_entry("symlink", "a link", None)
        entry.symlink_target = ""
        self.assertEqual(b"link\x00", inventory_delta._link_content(entry))

    def test_link_unicode_target(self):
        entry = inventory.make_entry("symlink", "a link", None)
        entry.symlink_target = b" \xc3\xa5".decode("utf8")
        self.assertEqual(b"link\x00 \xc3\xa5", inventory_delta._link_content(entry))

    def test_link_space_target(self):
        entry = inventory.make_entry("symlink", "a link", None)
        entry.symlink_target = " "
        self.assertEqual(b"link\x00 ", inventory_delta._link_content(entry))

    def test_link_no_target(self):
        entry = inventory.make_entry("symlink", "a link", None)
        self.assertRaises(InventoryDeltaError, inventory_delta._link_content, entry)

    def test_reference_null(self):
        entry = inventory.make_entry("tree-reference", "a tree", None)
        entry.reference_revision = NULL_REVISION
        self.assertEqual(b"tree\x00null:", inventory_delta._reference_content(entry))

    def test_reference_revision(self):
        entry = inventory.make_entry("tree-reference", "a tree", None)
        entry.reference_revision = b"foo@\xc3\xa5b-lah"
        self.assertEqual(
            b"tree\x00foo@\xc3\xa5b-lah", inventory_delta._reference_content(entry)
        )

    def test_reference_no_reference(self):
        entry = inventory.make_entry("tree-reference", "a tree", None)
        self.assertRaises(
            InventoryDeltaError, inventory_delta._reference_content, entry
        )



# Copyright (C) 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests for chunks_to_lines."""

from .. import tests
from . import features


def load_tests(loader, standard_tests, pattern):
    suite, _ = tests.permute_tests_for_extension(
        standard_tests,
        loader,
        "breezy._chunks_to_lines_py",
        "breezy._chunks_to_lines_pyx",
    )
    return suite


# test_osutils depends on this feature being around. We can't just use the one
# generated by load_tests, because if we only load osutils but not this module,
# then that code never gets run
compiled_chunkstolines_feature = features.ModuleAvailableFeature(
    "breezy._chunks_to_lines_pyx"
)


class TestChunksToLines(tests.TestCase):
    module = None  # Filled in by test parameterization

    def assertChunksToLines(self, lines, chunks, alreadly_lines=False):
        result = self.module.chunks_to_lines(chunks)
        self.assertEqual(lines, result)
        if alreadly_lines:
            self.assertIs(chunks, result)

    def test_fulltext_chunk_to_lines(self):
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz\n"], [b"foo\nbar\r\nba\rz\n"]
        )
        self.assertChunksToLines(
            [b"foobarbaz\n"], [b"foobarbaz\n"], alreadly_lines=True
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\n", b"\n", b"baz\n", b"\n", b"\n"],
            [b"foo\nbar\n\nbaz\n\n\n"],
        )
        self.assertChunksToLines([b"foobarbaz"], [b"foobarbaz"], alreadly_lines=True)
        self.assertChunksToLines([b"foobarbaz"], [b"foo", b"bar", b"baz"])

    def test_newlines(self):
        self.assertChunksToLines([b"\n"], [b"\n"], alreadly_lines=True)
        self.assertChunksToLines([b"\n"], [b"", b"\n", b""])
        self.assertChunksToLines([b"\n"], [b"\n", b""])
        self.assertChunksToLines([b"\n"], [b"", b"\n"])
        self.assertChunksToLines([b"\n", b"\n", b"\n"], [b"\n\n\n"])
        self.assertChunksToLines(
            [b"\n", b"\n", b"\n"], [b"\n", b"\n", b"\n"], alreadly_lines=True
        )

    def test_lines_to_lines(self):
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz\n"],
            [b"foo\n", b"bar\r\n", b"ba\rz\n"],
            alreadly_lines=True,
        )

    def test_no_final_newline(self):
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"], [b"foo\nbar\r\nba\rz"]
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"],
            [b"foo\n", b"bar\r\n", b"ba\rz"],
            alreadly_lines=True,
        )
        self.assertChunksToLines(
            (b"foo\n", b"bar\r\n", b"ba\rz"),
            (b"foo\n", b"bar\r\n", b"ba\rz"),
            alreadly_lines=True,
        )
        self.assertChunksToLines([], [], alreadly_lines=True)
        self.assertChunksToLines([b"foobarbaz"], [b"foobarbaz"], alreadly_lines=True)
        self.assertChunksToLines([], [b""])

    def test_mixed(self):
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"], [b"foo\n", b"bar\r\nba\r", b"z"]
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"], [b"foo\nb", b"a", b"r\r\nba\r", b"z"]
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"], [b"foo\nbar\r\nba", b"\r", b"z"]
        )

        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz"], [b"foo\n", b"", b"bar\r\nba", b"\r", b"z"]
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz\n"], [b"foo\n", b"bar\r\n", b"ba\rz\n", b""]
        )
        self.assertChunksToLines(
            [b"foo\n", b"bar\r\n", b"ba\rz\n"], [b"foo\n", b"bar", b"\r\n", b"ba\rz\n"]
        )

    def test_not_lines(self):
        # We should raise a TypeError, not crash
        self.assertRaises(TypeError, self.module.chunks_to_lines, object())
        self.assertRaises(TypeError, self.module.chunks_to_lines, [object()])
        self.assertRaises(TypeError, self.module.chunks_to_lines, [b"foo", object()])



# Copyright (C) 2006, 2007, 2009, 2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import tests


class TestRevisionHistory(tests.TestCaseWithTransport):
    def _build_branch(self):
        # setup a standalone branch with three commits
        tree = self.make_branch_and_tree("test")
        with open("test/foo", "wb") as f:
            f.write(b"1111\n")
        tree.add("foo")
        tree.commit("added foo", rev_id=b"revision_1")
        with open("test/foo", "wb") as f:
            f.write(b"2222\n")
        tree.commit("updated foo", rev_id=b"revision_2")
        with open("test/foo", "wb") as f:
            f.write(b"3333\n")
        tree.commit("updated foo again", rev_id=b"revision_3")
        return tree

    def _check_revision_history(self, location="", working_dir=None):
        rh = self.run_bzr(["revision-history", location], working_dir=working_dir)[0]
        self.assertEqual(rh, "revision_1\nrevision_2\nrevision_3\n")

    def test_revision_history(self):
        """No location."""
        self._build_branch()
        self._check_revision_history(working_dir="test")

    def test_revision_history_with_location(self):
        """With a specified location."""
        self._build_branch()
        self._check_revision_history("test")

    def test_revision_history_with_repo_branch(self):
        """With a repository branch location."""
        self._build_branch()
        self.run_bzr("init-shared-repo repo")
        self.run_bzr("branch test repo/test")
        self._check_revision_history("repo/test")

    def test_revision_history_with_checkout(self):
        """With a repository branch checkout location."""
        self._build_branch()
        self.run_bzr("init-shared-repo repo")
        self.run_bzr("branch test repo/test")
        self.run_bzr("checkout repo/test test-checkout")
        self._check_revision_history("test-checkout")

    def test_revision_history_with_lightweight_checkout(self):
        """With a repository branch lightweight checkout location."""
        self._build_branch()
        self.run_bzr("init-shared-repo repo")
        self.run_bzr("branch test repo/test")
        self.run_bzr("checkout --lightweight repo/test test-checkout")
        self._check_revision_history("test-checkout")



# Copyright (C) 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# cython: language_level=3

"""Functionality for doing annotations in the 'optimal' way"""


cdef extern from "python-compat.h":
    pass

from cpython.dict cimport PyDict_GetItem, PyDict_SetItem
from cpython.list cimport (PyList_Append, PyList_CheckExact, PyList_GET_ITEM,
                           PyList_GET_SIZE, PyList_SetItem, PyList_Sort)
from cpython.object cimport Py_EQ, Py_LT, PyObject, PyObject_RichCompareBool
from cpython.ref cimport Py_INCREF
from cpython.tuple cimport (PyTuple_CheckExact, PyTuple_GET_ITEM,
                            PyTuple_GET_SIZE, PyTuple_New, PyTuple_SET_ITEM)


cdef extern from "Python.h":
    void PyTuple_SET_ITEM_ptr "PyTuple_SET_ITEM" (object, Py_ssize_t,
                                                  PyObject *)

    void Py_INCREF_ptr "Py_INCREF" (PyObject *)
    void Py_DECREF_ptr "Py_DECREF" (PyObject *)

    int PyObject_RichCompareBool_ptr "PyObject_RichCompareBool" (
        PyObject *, PyObject *, int opid)


from . import _annotator_py


cdef int _check_match_ranges(list parent_annotations, list annotations,
                             Py_ssize_t parent_idx, Py_ssize_t lines_idx,
                             Py_ssize_t match_len) except -1:
    if parent_idx + match_len > len(parent_annotations):
        raise ValueError('Match length exceeds len of'
                         ' parent_annotations %s > %s'
                         % (parent_idx + match_len,
                            len(parent_annotations)))
    if lines_idx + match_len > len(annotations):
        raise ValueError('Match length exceeds len of'
                         ' annotations %s > %s'
                         % (lines_idx + match_len,
                            len(annotations)))
    return 0


cdef PyObject *_next_tuple_entry(object tpl, Py_ssize_t *pos): # cannot_raise
    """Return the next entry from this tuple.

    :param tpl: The tuple we are investigating, *must* be a PyTuple
    :param pos: The last item we found. Will be updated to the new position.

    This cannot raise an exception, as it does no error checking.
    """
    pos[0] = pos[0] + 1
    if pos[0] >= PyTuple_GET_SIZE(tpl):
        return NULL
    return PyTuple_GET_ITEM(tpl, pos[0])


cdef object _combine_annotations(ann_one, ann_two, cache):
    """Combine the annotations from both sides."""
    cdef Py_ssize_t pos_one, pos_two, len_one, len_two
    cdef Py_ssize_t out_pos
    cdef PyObject *temp
    cdef PyObject *left
    cdef PyObject *right

    if (PyObject_RichCompareBool(ann_one, ann_two, Py_LT)):
        cache_key = (ann_one, ann_two)
    else:
        cache_key = (ann_two, ann_one)
    temp = PyDict_GetItem(cache, cache_key)
    if temp != NULL:
        return <object>temp

    if not PyTuple_CheckExact(ann_one) or not PyTuple_CheckExact(ann_two):
        raise TypeError('annotations must be tuples')
    # We know that annotations are tuples, and that both sides are already
    # sorted, so we can just walk and update a new list.
    pos_one = -1
    pos_two = -1
    out_pos = 0
    left = _next_tuple_entry(ann_one, &pos_one)
    right = _next_tuple_entry(ann_two, &pos_two)
    new_ann = PyTuple_New(PyTuple_GET_SIZE(ann_one)
                          + PyTuple_GET_SIZE(ann_two))
    while left != NULL and right != NULL:
        # left == right is done by PyObject_RichCompareBool_ptr, however it
        # avoids a function call for a very common case. Drops 'time bzr
        # annotate NEWS' from 7.25s to 7.16s, so it *is* a visible impact.
        if (left == right
            or PyObject_RichCompareBool_ptr(left, right, Py_EQ)):
            # Identical values, step both
            Py_INCREF_ptr(left)
            PyTuple_SET_ITEM_ptr(new_ann, out_pos, left)
            left = _next_tuple_entry(ann_one, &pos_one)
            right = _next_tuple_entry(ann_two, &pos_two)
        elif (PyObject_RichCompareBool_ptr(left, right, Py_LT)):
            # left < right or right == NULL
            Py_INCREF_ptr(left)
            PyTuple_SET_ITEM_ptr(new_ann, out_pos, left)
            left = _next_tuple_entry(ann_one, &pos_one)
        else: # right < left or left == NULL
            Py_INCREF_ptr(right)
            PyTuple_SET_ITEM_ptr(new_ann, out_pos, right)
            right = _next_tuple_entry(ann_two, &pos_two)
        out_pos = out_pos + 1
    while left != NULL:
        Py_INCREF_ptr(left)
        PyTuple_SET_ITEM_ptr(new_ann, out_pos, left)
        left = _next_tuple_entry(ann_one, &pos_one)
        out_pos = out_pos + 1
    while right != NULL:
        Py_INCREF_ptr(right)
        PyTuple_SET_ITEM_ptr(new_ann, out_pos, right)
        right = _next_tuple_entry(ann_two, &pos_two)
        out_pos = out_pos + 1
    if out_pos != PyTuple_GET_SIZE(new_ann):
        # Timing _PyTuple_Resize was not significantly faster that slicing
        # PyTuple_Resize((<PyObject **>new_ann), out_pos)
        new_ann = new_ann[0:out_pos]
    PyDict_SetItem(cache, cache_key, new_ann)
    return new_ann


cdef int _apply_parent_annotations(list annotations, list parent_annotations,
                                   matching_blocks) except -1:
    """Apply the annotations from parent_annotations into annotations.

    matching_blocks defines the ranges that match.
    """
    cdef Py_ssize_t parent_idx, lines_idx, match_len, idx
    cdef PyObject *par_temp
    cdef PyObject *ann_temp

    # For NEWS and breezy/builtins.py, over 99% of the lines are simply copied
    # across from the parent entry. So this routine is heavily optimized for
    # that. Would be interesting if we could use memcpy() but we have to incref
    # and decref
    for parent_idx, lines_idx, match_len in matching_blocks:
        _check_match_ranges(parent_annotations, annotations,
                            parent_idx, lines_idx, match_len)
        annotations[lines_idx:lines_idx + match_len] = parent_annotations[parent_idx:parent_idx + match_len]
    return 0


cdef int _merge_annotations(object this_annotation, list annotations,
                            list parent_annotations,
                            matching_blocks, ann_cache) except -1:
    cdef Py_ssize_t parent_idx, ann_idx, lines_idx, match_len, idx
    cdef Py_ssize_t pos
    cdef object ann
    cdef object par

    last_ann = None
    last_parent = None
    last_res = None
    for parent_idx, lines_idx, match_len in matching_blocks:
        _check_match_ranges(parent_annotations, annotations,
                            parent_idx, lines_idx, match_len)
        # For lines which match this parent, we will now resolve whether
        # this parent wins over the current annotation
        for idx from 0 <= idx < match_len:
            ann_idx = lines_idx + idx
            ann = annotations[ann_idx]
            par_ann = parent_annotations[parent_idx + idx]
            if (ann == par_ann):
                # This is parent, do nothing
                # Pointer comparison is fine here. Value comparison would
                # be ok, but it will be handled in the final if clause by
                # merging the two tuples into the same tuple
                # Avoiding the Py_INCREF and function call to
                # PyObject_RichCompareBool using pointer comparison drops
                # timing from 215ms => 125ms
                continue
            if (ann is this_annotation):
                # Originally claimed 'this', but it was really in this
                # parent
                annotations[ann_idx] = par_ann
                continue
            # Resolve the fact that both sides have a different value for
            # last modified
            if (ann is last_ann and par_ann is last_parent):
                annotations[ann_idx] = last_res
            else:
                new_ann = _combine_annotations(ann, par_ann, ann_cache)
                annotations[ann_idx] = new_ann
                last_ann = ann
                last_parent = par_ann
                last_res = new_ann
    return 0


class Annotator(_annotator_py.Annotator):
    """Class that drives performing annotations."""

    def _update_from_first_parent(self, key, annotations, lines, parent_key):
        """Reannotate this text relative to its first parent."""
        (parent_annotations,
         matching_blocks) = self._get_parent_annotations_and_matches(
                                key, lines, parent_key)

        _apply_parent_annotations(annotations, parent_annotations,
                                  matching_blocks)

    def _update_from_other_parents(self, key, annotations, lines,
                                   this_annotation, parent_key):
        """Reannotate this text relative to a second (or more) parent."""
        (parent_annotations,
         matching_blocks) = self._get_parent_annotations_and_matches(
                                key, lines, parent_key)
        _merge_annotations(this_annotation, annotations, parent_annotations,
                           matching_blocks, self._ann_tuple_cache)

    def annotate_flat(self, key):
        """Determine the single-best-revision to source for each line.

        This is meant as a compatibility thunk to how annotate() used to work.
        """
        cdef Py_ssize_t pos, num_lines

        from . import annotate

        custom_tiebreaker = annotate._break_annotation_tie
        annotations, lines = self.annotate(key)
        num_lines = len(lines)
        out = []
        heads = self._get_heads_provider().heads
        for pos from 0 <= pos < num_lines:
            annotation = annotations[pos]
            line = lines[pos]
            if len(annotation) == 1:
                head = annotation[0]
            else:
                the_heads = heads(annotation)
                if len(the_heads) == 1:
                    for head in the_heads: break # get the item out of the set
                else:
                    # We need to resolve the ambiguity, for now just pick the
                    # sorted smallest
                    head = self._resolve_annotation_tie(the_heads, line,
                                                        custom_tiebreaker)
            PyList_Append(out, (head, line))
        return out



# Copyright (C) 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for repository implementations - tests a repository format."""

from breezy import errors, gpg, tests
from breezy import repository as _mod_repository
from breezy import revision as _mod_revision
from breezy.bzr import inventory, versionedfile, vf_repository
from breezy.bzr.tests.per_repository_vf import (
    TestCaseWithRepository,
    all_repository_vf_format_scenarios,
)
from breezy.tests.matchers import MatchesAncestry
from breezy.tests.scenarios import load_tests_apply_scenarios

load_tests = load_tests_apply_scenarios


class TestRepository(TestCaseWithRepository):
    scenarios = all_repository_vf_format_scenarios()

    def assertFormatAttribute(self, attribute, allowed_values):
        """Assert that the format has an attribute 'attribute'."""
        repo = self.make_repository("repo")
        self.assertSubset([getattr(repo._format, attribute)], allowed_values)

    def test_attribute__fetch_order(self):
        """Test the _fetch_order attribute."""
        self.assertFormatAttribute("_fetch_order", ("topological", "unordered"))

    def test_attribute__fetch_uses_deltas(self):
        """Test the _fetch_uses_deltas attribute."""
        self.assertFormatAttribute("_fetch_uses_deltas", (True, False))

    def test_attribute_inventories_store(self):
        """Test the existence of the inventories attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        self.assertIsInstance(repo.inventories, versionedfile.VersionedFiles)

    def test_attribute_inventories_basics(self):
        """Test basic aspects of the inventories attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        rev_id = (tree.commit("a"),)
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual({rev_id}, set(repo.inventories.keys()))

    def test_attribute_revision_store(self):
        """Test the existence of the revisions attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        self.assertIsInstance(repo.revisions, versionedfile.VersionedFiles)

    def test_attribute_revision_store_basics(self):
        """Test the basic behaviour of the revisions attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        repo.lock_write()
        try:
            self.assertEqual(set(), set(repo.revisions.keys()))
            revid = (tree.commit("foo"),)
            self.assertEqual({revid}, set(repo.revisions.keys()))
            self.assertEqual({revid: ()}, repo.revisions.get_parent_map([revid]))
        finally:
            repo.unlock()
        tree2 = self.make_branch_and_tree("tree2")
        tree2.pull(tree.branch)
        left_id = (tree2.commit("left"),)
        right_id = (tree.commit("right"),)
        tree.merge_from_branch(tree2.branch)
        merge_id = (tree.commit("merged"),)
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(
            {revid, left_id, right_id, merge_id}, set(repo.revisions.keys())
        )
        self.assertEqual(
            {
                revid: (),
                left_id: (revid,),
                right_id: (revid,),
                merge_id: (right_id, left_id),
            },
            repo.revisions.get_parent_map(repo.revisions.keys()),
        )

    def test_attribute_signature_store(self):
        """Test the existence of the signatures attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        self.assertIsInstance(repo.signatures, versionedfile.VersionedFiles)

    def test_exposed_versioned_files_are_marked_dirty(self):
        repo = self.make_repository(".")
        repo.lock_write()
        signatures = repo.signatures
        revisions = repo.revisions
        inventories = repo.inventories
        repo.unlock()
        self.assertRaises(errors.ObjectNotLocked, signatures.keys)
        self.assertRaises(errors.ObjectNotLocked, revisions.keys)
        self.assertRaises(errors.ObjectNotLocked, inventories.keys)
        self.assertRaises(
            errors.ObjectNotLocked, signatures.add_lines, ("foo",), [], []
        )
        self.assertRaises(errors.ObjectNotLocked, revisions.add_lines, ("foo",), [], [])
        self.assertRaises(
            errors.ObjectNotLocked, inventories.add_lines, ("foo",), [], []
        )

    def test__get_sink(self):
        repo = self.make_repository("repo")
        sink = repo._get_sink()
        self.assertIsInstance(sink, vf_repository.StreamSink)

    def test_get_serializer_format(self):
        repo = self.make_repository(".")
        format = repo.get_serializer_format()
        self.assertEqual(repo._serializer.format_num, format)

    def test_add_revision_inventory_sha1(self):
        inv = inventory.Inventory(revision_id=b"A")
        inv.root.revision = b"A"
        inv.root.file_id = b"fixed-root"
        # Insert the inventory on its own to an identical repository, to get
        # its sha1.
        reference_repo = self.make_repository("reference_repo")
        reference_repo.lock_write()
        reference_repo.start_write_group()
        inv_sha1 = reference_repo.add_inventory(b"A", inv, [])
        reference_repo.abort_write_group()
        reference_repo.unlock()
        # Now insert a revision with this inventory, and it should get the same
        # sha1.
        repo = self.make_repository("repo")
        repo.lock_write()
        repo.start_write_group()
        repo.texts.add_lines((b"fixed-root", b"A"), [], [])
        repo.add_revision(
            b"A",
            _mod_revision.Revision(
                b"A", committer="B", timestamp=0, timezone=0, message="C"
            ),
            inv=inv,
        )
        repo.commit_write_group()
        repo.unlock()
        repo.lock_read()
        self.assertEqual(inv_sha1, repo.get_revision(b"A").inventory_sha1)
        repo.unlock()

    def test_install_revisions(self):
        wt = self.make_branch_and_tree("source")
        wt.commit("A", allow_pointless=True, rev_id=b"A")
        repo = wt.branch.repository
        repo.lock_write()
        repo.start_write_group()
        repo.sign_revision(b"A", gpg.LoopbackGPGStrategy(None))
        repo.commit_write_group()
        repo.unlock()
        repo.lock_read()
        self.addCleanup(repo.unlock)
        repo2 = self.make_repository("repo2")
        revision = repo.get_revision(b"A")
        tree = repo.revision_tree(b"A")
        signature = repo.get_signature_text(b"A")
        repo2.lock_write()
        self.addCleanup(repo2.unlock)
        vf_repository.install_revisions(repo2, [(revision, tree, signature)])
        self.assertEqual(revision, repo2.get_revision(b"A"))
        self.assertEqual(signature, repo2.get_signature_text(b"A"))

    def test_attribute_text_store(self):
        """Test the existence of the texts attribute."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        self.assertIsInstance(repo.texts, versionedfile.VersionedFiles)

    def test_iter_inventories_is_ordered(self):
        # just a smoke test
        tree = self.make_branch_and_tree("a")
        first_revision = tree.commit("")
        second_revision = tree.commit("")
        tree.lock_read()
        self.addCleanup(tree.unlock)
        revs = (first_revision, second_revision)
        invs = tree.branch.repository.iter_inventories(revs)
        for rev_id, inv in zip(revs, invs):
            self.assertEqual(rev_id, inv.revision_id)
            self.assertIsInstance(inv, inventory.CommonInventory)

    def test_item_keys_introduced_by(self):
        # Make a repo with one revision and one versioned file.
        tree = self.make_branch_and_tree("t")
        self.build_tree(["t/foo"])
        tree.add("foo", ids=b"file1")
        tree.commit("message", rev_id=b"rev_id")
        repo = tree.branch.repository
        repo.lock_write()
        repo.start_write_group()
        try:
            repo.sign_revision(b"rev_id", gpg.LoopbackGPGStrategy(None))
        except errors.UnsupportedOperation:
            signature_texts = []
        else:
            signature_texts = [b"rev_id"]
        repo.commit_write_group()
        repo.unlock()
        repo.lock_read()
        self.addCleanup(repo.unlock)

        # Item keys will be in this order, for maximum convenience for
        # generating data to insert into knit repository:
        #   * files
        #   * inventory
        #   * signatures
        #   * revisions
        expected_item_keys = [
            ("file", b"file1", [b"rev_id"]),
            ("inventory", None, [b"rev_id"]),
            ("signatures", None, signature_texts),
            ("revisions", None, [b"rev_id"]),
        ]
        item_keys = list(repo.item_keys_introduced_by([b"rev_id"]))
        item_keys = [
            (kind, file_id, list(versions)) for (kind, file_id, versions) in item_keys
        ]

        if repo.supports_rich_root():
            # Check for the root versioned file in the item_keys, then remove
            # it from streamed_names so we can compare that with
            # expected_record_names.
            # Note that the file keys can be in any order, so this test is
            # written to allow that.
            inv = repo.get_inventory(b"rev_id")
            root_item_key = ("file", inv.root.file_id, [b"rev_id"])
            self.assertIn(root_item_key, item_keys)
            item_keys.remove(root_item_key)

        self.assertEqual(expected_item_keys, item_keys)

    def test_attribute_text_store_basics(self):
        """Test the basic behaviour of the text store."""
        tree = self.make_branch_and_tree("tree")
        repo = tree.branch.repository
        file_id = b"Foo:Bar"
        file_key = (file_id,)
        with tree.lock_write():
            self.assertEqual(set(), set(repo.texts.keys()))
            tree.add(["foo"], ["file"], [file_id])
            tree.put_file_bytes_non_atomic("foo", b"content\n")
            try:
                rev_key = (tree.commit("foo"),)
            except errors.IllegalPath as e:
                raise tests.TestNotApplicable(
                    "file_id {!r} cannot be stored on this"
                    " platform for this repo format".format(file_id)
                ) from e
            if repo._format.rich_root_data:
                root_commit = (tree.path2id(""),) + rev_key
                keys = {root_commit}
                parents = {root_commit: ()}
            else:
                keys = set()
                parents = {}
            keys.add(file_key + rev_key)
            parents[file_key + rev_key] = ()
            self.assertEqual(keys, set(repo.texts.keys()))
            self.assertEqual(parents, repo.texts.get_parent_map(repo.texts.keys()))
        tree2 = self.make_branch_and_tree("tree2")
        tree2.pull(tree.branch)
        tree2.put_file_bytes_non_atomic("foo", b"right\n")
        right_key = (tree2.commit("right"),)
        keys.add(file_key + right_key)
        parents[file_key + right_key] = (file_key + rev_key,)
        tree.put_file_bytes_non_atomic("foo", b"left\n")
        left_key = (tree.commit("left"),)
        keys.add(file_key + left_key)
        parents[file_key + left_key] = (file_key + rev_key,)
        tree.merge_from_branch(tree2.branch)
        tree.put_file_bytes_non_atomic("foo", b"merged\n")
        try:
            tree.auto_resolve()
        except errors.UnsupportedOperation:
            pass
        merge_key = (tree.commit("merged"),)
        keys.add(file_key + merge_key)
        parents[file_key + merge_key] = (file_key + left_key, file_key + right_key)
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual(keys, set(repo.texts.keys()))
        self.assertEqual(parents, repo.texts.get_parent_map(repo.texts.keys()))


class TestCaseWithComplexRepository(TestCaseWithRepository):
    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()
        tree_a = self.make_branch_and_tree("a")
        self.controldir = tree_a.branch.controldir
        # add a corrupt inventory 'orphan'
        # this may need some generalising for knits.
        with tree_a.lock_write(), _mod_repository.WriteGroup(tree_a.branch.repository):
            inv_file = tree_a.branch.repository.inventories
            inv_file.add_lines((b"orphan",), [], [])
        # add a real revision 'rev1'
        tree_a.commit("rev1", rev_id=b"rev1", allow_pointless=True)
        # add a real revision 'rev2' based on rev1
        tree_a.commit("rev2", rev_id=b"rev2", allow_pointless=True)
        # add a reference to a ghost
        tree_a.add_parent_tree_id(b"ghost1")
        try:
            tree_a.commit("rev3", rev_id=b"rev3", allow_pointless=True)
        except errors.RevisionNotPresent as e:
            raise tests.TestNotApplicable(
                "Cannot test with ghosts for this format."
            ) from e
        # add another reference to a ghost, and a second ghost.
        tree_a.add_parent_tree_id(b"ghost1")
        tree_a.add_parent_tree_id(b"ghost2")
        tree_a.commit("rev4", rev_id=b"rev4", allow_pointless=True)

    def test_revision_trees(self):
        revision_ids = [b"rev1", b"rev2", b"rev3", b"rev4"]
        repository = self.controldir.open_repository()
        repository.lock_read()
        self.addCleanup(repository.unlock)
        trees1 = list(repository.revision_trees(revision_ids))
        trees2 = [repository.revision_tree(t) for t in revision_ids]
        self.assertEqual(len(trees1), len(trees2))
        for tree1, tree2 in zip(trees1, trees2):
            self.assertFalse(tree2.changes_from(tree1).has_changed())

    def test_get_revision_deltas(self):
        repository = self.controldir.open_repository()
        repository.lock_read()
        self.addCleanup(repository.unlock)
        revisions = [
            repository.get_revision(r) for r in [b"rev1", b"rev2", b"rev3", b"rev4"]
        ]
        deltas1 = list(repository.get_revision_deltas(revisions))
        deltas2 = [repository.get_revision_delta(r.revision_id) for r in revisions]
        self.assertEqual(deltas1, deltas2)

    def test_all_revision_ids(self):
        # all_revision_ids -> all revisions
        self.assertEqual(
            {b"rev1", b"rev2", b"rev3", b"rev4"},
            set(self.controldir.open_repository().all_revision_ids()),
        )

    def test_reserved_id(self):
        repo = self.make_repository("repository")
        with repo.lock_write(), _mod_repository.WriteGroup(repo):
            self.assertRaises(
                errors.ReservedId, repo.add_inventory, b"reserved:", None, None
            )
            self.assertRaises(
                errors.ReservedId,
                repo.add_inventory_by_delta,
                "foo",
                [],
                b"reserved:",
                None,
            )
            self.assertRaises(errors.ReservedId, repo.add_revision, b"reserved:", None)


class TestCaseWithCorruptRepository(TestCaseWithRepository):
    scenarios = all_repository_vf_format_scenarios()

    def setUp(self):
        super().setUp()
        # a inventory with no parents and the revision has parents..
        # i.e. a ghost.
        repo = self.make_repository("inventory_with_unnecessary_ghost")
        repo.lock_write()
        repo.start_write_group()
        inv = inventory.Inventory(revision_id=b"ghost")
        inv.root.revision = b"ghost"
        if repo.supports_rich_root():
            root_id = inv.root.file_id
            repo.texts.add_lines((root_id, b"ghost"), [], [])
        sha1 = repo.add_inventory(b"ghost", inv, [])
        rev = _mod_revision.Revision(
            timestamp=0,
            timezone=None,
            committer="Foo Bar <foo@example.com>",
            message="Message",
            inventory_sha1=sha1,
            revision_id=b"ghost",
        )
        rev.parent_ids = [b"the_ghost"]
        try:
            repo.add_revision(b"ghost", rev)
        except (errors.NoSuchRevision, errors.RevisionNotPresent) as e:
            raise tests.TestNotApplicable(
                "Cannot test with ghosts for this format."
            ) from e

        inv = inventory.Inventory(revision_id=b"the_ghost")
        inv.root.revision = b"the_ghost"
        if repo.supports_rich_root():
            root_id = inv.root.file_id
            repo.texts.add_lines((root_id, b"the_ghost"), [], [])
        sha1 = repo.add_inventory(b"the_ghost", inv, [])
        rev = _mod_revision.Revision(
            timestamp=0,
            timezone=None,
            committer="Foo Bar <foo@example.com>",
            message="Message",
            inventory_sha1=sha1,
            revision_id=b"the_ghost",
        )
        rev.parent_ids = []
        repo.add_revision(b"the_ghost", rev)
        # check its setup usefully
        inv_weave = repo.inventories
        possible_parents = (None, ((b"ghost",),))
        self.assertSubset(
            inv_weave.get_parent_map([(b"ghost",)])[(b"ghost",)], possible_parents
        )
        repo.commit_write_group()
        repo.unlock()

    def test_corrupt_revision_access_asserts_if_reported_wrong(self):
        repo_url = self.get_url("inventory_with_unnecessary_ghost")
        repo = _mod_repository.Repository.open(repo_url)
        m = MatchesAncestry(repo, b"ghost")
        reported_wrong = False
        try:
            if m.match([b"the_ghost", b"ghost"]) is not None:
                reported_wrong = True
        except errors.CorruptRepository:
            # caught the bad data:
            return
        if not reported_wrong:
            return
        self.assertRaises(errors.CorruptRepository, repo.get_revision, b"ghost")

    def test_corrupt_revision_get_revision_reconcile(self):
        repo_url = self.get_url("inventory_with_unnecessary_ghost")
        repo = _mod_repository.Repository.open(repo_url)
        repo.get_revision_reconcile(b"ghost")
